@@ -69,8 +69,16 @@ swfmoz_player_menu_about (GtkMenuItem *item, SwfmozPlayer *player)
     "Eric Anholt <eric@anholt.net>",
     NULL,
   };
+  static const char *artists[] = {
+    "Andreas Nilsson <andreas@andreasn.se>",
+    "Cristian Grada <krigenator@gmail.com>",
+    NULL
+  };
+
   gtk_show_about_dialog (NULL, "program-name", "Swfdec",
+      "logo-icon-name", "swfdec",
       "authors", authors,
+      "artists", artists,
       "comments", "Play Flash content in your browser",
       "name", "Swfdec Mozilla Plugin",
       "version", VERSION,
@@ -138,65 +146,89 @@ static gboolean
 swfmoz_player_idle_redraw (gpointer playerp)
 {
   SwfmozPlayer *player = playerp;
+  GdkRegion *region;
+
+  region = player->repaint;
+  if (region) {
+    player->repaint = NULL;
+    g_source_destroy (player->repaint_source);
+    g_source_unref (player->repaint_source);
+    player->repaint_source = NULL;
+    swfmoz_player_render (player, region);
+    gdk_region_destroy (region);
+  }
 
-  swfmoz_player_render (player, player->x, player->y, player->width, player->height);
   return TRUE;
 }
 
 static void
-swfmoz_player_redraw (SwfdecPlayer *swfplayer, double x, double y, 
-    double width, double height, SwfmozPlayer *player)
+swfmoz_player_redraw (SwfdecPlayer *swfplayer, const SwfdecRectangle *extents, 
+    const SwfdecRectangle *rects, guint n_rects, SwfmozPlayer *player)
 {
-  int xi, yi, wi, hi;
+  GdkRegion *region;
+  guint i;
 
-  xi = MAX ((int) floor (x), 0);
-  yi = MAX ((int) floor (y), 0);
-  wi = (int) ceil (x + width) - xi;
-  hi = (int) ceil (y + height) - yi;
-  if (wi <= 0 || hi <= 0)
-    return;
+  if (player->repaint)
+    region = player->repaint;
+  else
+    region = gdk_region_new ();
+
+  for (i = 0; i < n_rects; i++) {
+    gdk_region_union_with_rect (region, (GdkRectangle *) &rects[i]);
+  }
 
-  xi += player->target_rect.x;
-  yi += player->target_rect.y;
   if (player->windowless) {
     NPRect rect;
-    rect.left = xi;
-    rect.top = yi;
-    rect.right = xi + wi;
-    rect.bottom = yi + hi;
-    plugin_invalidate_rect (player->instance, &rect);
-  } else if (player->repaint_source) {
-    player->width = MAX (player->x + player->width, xi + wi);
-    player->height = MAX (player->y + player->height, yi + hi);
-    player->x = MIN (player->x, xi);
-    player->y = MIN (player->y, yi);
-    player->width -= player->x;
-    player->height -= player->y;
+    GdkRectangle *rectangles;
+    int n_rectangles;
+
+    g_assert (player->repaint == NULL);
+    
+    gdk_region_get_rectangles (region, &rectangles, &n_rectangles);
+
+    for (i = 0; i < (guint) n_rectangles; i++) {
+      rect.left = rectangles[i].x;
+      rect.top = rectangles[i].y;
+      rect.right = rectangles[i].x + rectangles[i].width;
+      rect.bottom = rectangles[i].y + rectangles[i].height;
+      plugin_invalidate_rect (player->instance, &rect);
+    }
+    gdk_region_destroy (region);
   } else {
-    GSource *source = g_idle_source_new ();
-    player->repaint_source = source;
-    g_source_set_priority (source, GDK_PRIORITY_REDRAW);
-    g_source_set_callback (source, swfmoz_player_idle_redraw, player, NULL);
-    g_source_attach (source, player->context);
-    player->x = xi;
-    player->y = yi;
-    player->width = wi;
-    player->height = hi;
+    if (player->repaint_source) {
+      g_assert (player->repaint);
+    } else {
+      GSource *source = g_idle_source_new ();
+      player->repaint_source = source;
+      g_source_set_priority (source, GDK_PRIORITY_REDRAW);
+      g_source_set_callback (source, swfmoz_player_idle_redraw, player, NULL);
+      g_source_attach (source, player->context);
+      player->repaint = region;
+    }
   }
 }
 
 static void
-swfmoz_player_launch (SwfdecPlayer *swfplayer, const char *url, const char *target,
-    SwfmozPlayer *player)
+swfmoz_player_launch (SwfdecPlayer *swfplayer, SwfdecLoaderRequest request, 
+    const char *url, const char *target, SwfdecBuffer *data, SwfmozPlayer *player)
 {
-  plugin_get_url (player->instance, url, target);
+  if (request == SWFDEC_LOADER_REQUEST_POST) {
+    if (data) {
+      plugin_post_url (player->instance, url, target, 
+	  (const char *) data->data, data->length);
+    } else {
+      plugin_post_url (player->instance, url, target, NULL, 0);
+    }
+  } else {
+    plugin_get_url (player->instance, url, target);
+  }
 }
 
 static void
 swfmoz_player_invalidate (SwfmozPlayer *player)
 {
-  swfmoz_player_redraw (player->player, 0.0, 0.0, player->target_rect.width, 
-      player->target_rect.height, player);
+  SwfdecRectangle rect = { 0, 0, player->target_rect.width, player->target_rect.height };
+  swfmoz_player_redraw (player->player, &rect, &rect, 1, player);
 }
 
 /* function stolen from SwfdecGtkWidget */
@@ -313,6 +345,8 @@ swfmoz_player_dispose (GObject *object)
     g_source_destroy (player->repaint_source);
     g_source_unref (player->repaint_source);
     player->repaint_source = NULL;
+    gdk_region_destroy (player->repaint);
+    player->repaint = NULL;
   }
   if (player->initial) {
     g_object_unref (player->initial);
@@ -340,7 +374,7 @@ swfmoz_player_class_init (SwfmozPlayerClass *klass)
 static void
 swfmoz_player_init (SwfmozPlayer *player)
 {
-  player->player = swfdec_gtk_player_new ();
+  player->player = swfdec_gtk_player_new (NULL);
   g_signal_connect (player->player, "invalidate", G_CALLBACK (swfmoz_player_redraw), player);
   g_signal_connect (player->player, "launch", G_CALLBACK (swfmoz_player_launch), player);
   g_signal_connect (player->player, "notify", G_CALLBACK (swfmoz_player_notify), player);
@@ -367,22 +401,29 @@ static void
 swfmoz_player_loaders_update (GtkListStore *store, GtkTreeIter *iter, SwfdecLoader *loader)
 {
   char *filename = swfdec_loader_get_filename (loader);
-  guint percent;
+  glong percent;
+  gboolean eof;
 
   percent = swfdec_loader_get_size (loader);
-  if (percent) {
-    percent = 100 * swfdec_loader_get_loaded (loader) / percent;
-  } else {
+  if (percent == 0) {
+    percent = 100;
+  } else if (percent < 0) {
     percent = 50;
+  } else {
+    percent = 100 * swfdec_loader_get_loaded (loader) / percent;
+    percent = CLAMP (percent, 0, 100);
   }
+  /* FIXME: swfdec needs a function for this */
+  g_object_get (G_OBJECT (loader), "eof", &eof, NULL);
+
   gtk_list_store_set (store, iter,
     SWFMOZ_LOADER_COLUMN_LOADER, loader,
     SWFMOZ_LOADER_COLUMN_NAME, filename,
     SWFMOZ_LOADER_COLUMN_URL, swfdec_url_get_url (swfdec_loader_get_url (loader)),
-    SWFMOZ_LOADER_COLUMN_EOF, percent == 100,
+    SWFMOZ_LOADER_COLUMN_EOF, eof,
     SWFMOZ_LOADER_COLUMN_ERROR, loader->error != NULL,
     SWFMOZ_LOADER_COLUMN_TYPE, swfmoz_loader_get_data_type_string (loader),
-    SWFMOZ_LOADER_COLUMN_PERCENT_LOADED, percent,
+    SWFMOZ_LOADER_COLUMN_PERCENT_LOADED, (guint) percent,
     -1);
   g_free (filename);
 }
@@ -447,7 +488,7 @@ swfmoz_player_set_target (SwfmozPlayer *player, GdkWindow *target,
     int x, int y, int width, int height)
 {
   g_return_if_fail (SWFMOZ_IS_PLAYER (player));
-  g_return_if_fail (GDK_IS_WINDOW (target));
+  g_return_if_fail (target == NULL || GDK_IS_WINDOW (target));
 
   if (player->target) {
     g_object_unref (player->target);
@@ -464,96 +505,184 @@ swfmoz_player_set_target (SwfmozPlayer *player, GdkWindow *target,
   }
 }
 
+static void
+swfdec_gtk_player_draw_pause (cairo_t *cr)
+{
+  cairo_pattern_t *pattern;
+#if 0
+#define ALPHA 0.447059
+#else
+#define ALPHA 0.6
+#endif
+
+  cairo_set_line_width (cr, 0.959352);
+  cairo_set_miter_limit (cr, 4);
+  cairo_set_line_cap (cr, 0);
+  cairo_set_line_join (cr, 0);
+  cairo_move_to (cr, 16.0003, 0.567675);
+  cairo_curve_to (cr, 7.44305, 0.567675, 0.505822, 7.53349, 0.505822, 16.1256);
+  cairo_curve_to (cr, 0.505822, 24.7178, 7.44305, 31.6836, 16.0003, 31.6836);
+  cairo_curve_to (cr, 24.5576, 31.6836, 31.4948, 24.7178, 31.4948, 16.1256);
+  cairo_curve_to (cr, 31.4948, 7.53349, 24.5576, 0.567675, 16.0003, 0.567675);
+  cairo_close_path (cr);
+  cairo_move_to (cr, 16.0003, 0.567675);
+  cairo_move_to (cr, 16.0607, 3.36325);
+  cairo_curve_to (cr, 22.984, 3.36325, 28.5953, 8.99503, 28.5953, 15.9433);
+  cairo_curve_to (cr, 28.5953, 22.8916, 22.984, 28.5234, 16.0607, 28.5234);
+  cairo_curve_to (cr, 9.13743, 28.5234, 3.49599, 22.8916, 3.49599, 15.9433);
+  cairo_curve_to (cr, 3.49599, 8.99503, 9.13743, 3.36325, 16.0607, 3.36325);
+  cairo_close_path (cr);
+  cairo_move_to (cr, 16.0607, 3.36325);
+  cairo_set_fill_rule (cr, 0);
+  cairo_set_source_rgba (cr, 0.827451, 0.843137, 0.811765, ALPHA);
+  cairo_fill_preserve (cr);
+  pattern = cairo_pattern_create_linear (24.906, 26.4817, 3.6134, 5.18912);
+  cairo_pattern_add_color_stop_rgba (pattern, 0, 0, 0, 0, ALPHA);
+  cairo_pattern_add_color_stop_rgba (pattern, 1, 1, 1, 1, ALPHA);
+  cairo_set_source (cr, pattern);
+  cairo_stroke (cr);
+  cairo_set_line_width (cr, 1);
+  cairo_set_miter_limit (cr, 4);
+  cairo_move_to (cr, 11.4927, 8.57249);
+  cairo_line_to (cr, 23.5787, 16.0226);
+  cairo_line_to (cr, 11.6273, 23.4948);
+  cairo_line_to (cr, 11.4927, 8.57249);
+  cairo_close_path (cr);
+  cairo_move_to (cr, 11.4927, 8.57249);
+  cairo_set_fill_rule (cr, 1);
+  cairo_set_source_rgba (cr, 0.827451, 0.843137, 0.811765, ALPHA);
+  cairo_fill_preserve (cr);
+  cairo_set_source (cr, pattern);
+  cairo_stroke (cr);
+#if 0
+  cairo_scale(cr, 1.18644, 1.20339);
+  cairo_translate (cr, -5.3063, -2.93549);
+  cairo_set_line_width (cr, 0.8369);
+  cairo_set_miter_limit (cr, 4);
+  cairo_move_to (cr, 28.461, 15.7365);
+  cairo_curve_to (cr, 28.4623, 21.4977, 23.7923, 26.1687, 18.0312, 26.1687);
+  cairo_curve_to (cr, 12.2701, 26.1687, 7.60014, 21.4977, 7.6014, 15.7365);
+  cairo_curve_to (cr, 7.60014, 9.97542, 12.2701, 5.30444, 18.0312, 5.30444);
+  cairo_curve_to (cr, 23.7923, 5.30444, 28.4623, 9.97542, 28.461, 15.7365);
+  cairo_close_path (cr);
+  cairo_move_to (cr, 28.461, 15.7365);
+  cairo_set_source (cr, pattern);
+  cairo_stroke (cr);
+#endif
+  cairo_pattern_destroy (pattern);
+}
+
 void
-swfmoz_player_render (SwfmozPlayer *player, int x, int y, int width, int height)
+swfmoz_player_render (SwfmozPlayer *player, GdkRegion *region)
 {
   GdkRectangle rect;
   cairo_t *cr;
 
   g_return_if_fail (SWFMOZ_IS_PLAYER (player));
-  g_return_if_fail (x >= 0);
-  g_return_if_fail (y >= 0);
-  g_return_if_fail (width > 0);
-  g_return_if_fail (height > 0);
+  g_return_if_fail (!gdk_region_empty (region));
 
-  /* first, remove the idle source */
-  if (player->repaint_source && x <= player->x && y <= player->y &&
-      x + width >= player->x + player->width &&
-      y + height >= player->y + player->height) {
-    g_source_destroy (player->repaint_source);
-    g_source_unref (player->repaint_source);
-    player->repaint_source = NULL;
+  /* first, remove the repainted stuff from the stuff that needs a redraw */
+  if (player->repaint) {
+
+    g_assert (player->repaint_source);
+    gdk_region_subtract (player->repaint, region);
+
+    if (gdk_region_empty (player->repaint)) {
+      g_source_destroy (player->repaint_source);
+      g_source_unref (player->repaint_source);
+      player->repaint_source = NULL;
+      gdk_region_destroy (player->repaint);
+      player->repaint = NULL;
+    }
   }
 
   /* second, check if we have anything to draw */
   if (player->target == NULL)
     return;
-  rect.x = x + player->target_rect.x;
-  rect.y = y + player->target_rect.y;
-  rect.width = width;
-  rect.height = height;
 
   /* paint it */
-  gdk_window_begin_paint_rect (player->target, &rect);
+  gdk_window_begin_paint_region (player->target, region);
   cr = gdk_cairo_create (player->target);
+  gdk_cairo_region (cr, region);
+  cairo_clip (cr);
   cairo_translate (cr, player->target_rect.x, player->target_rect.y);
-  swfdec_player_render (player->player, cr, x, y, rect.width, rect.height);
+  gdk_region_get_clipbox (region, &rect);
+  swfdec_player_render (player->player, cr, rect.x - player->target_rect.x,
+      rect.y - player->target_rect.y, rect.width, rect.height);
   /* paint optional pause sign */
   if (!swfdec_gtk_player_get_playing (SWFDEC_GTK_PLAYER (player->player))) {
     int w = player->target_rect.width;
     int h = player->target_rect.height;
     int len = MIN (w, h) * 4 / 5;
-    cairo_set_fill_rule (cr, CAIRO_FILL_RULE_EVEN_ODD);
-    cairo_set_source_rgba (cr, 1.0, 1.0, 1.0, 0.6);
-    cairo_rectangle (cr, (w - len) / 2, (h - len) / 2,
-	len / 3, len);
-    cairo_rectangle (cr, (w - len) / 2 + 2 * len / 3, (h - len) / 2,
-	len / 3, len);
-    cairo_fill_preserve (cr);
     cairo_rectangle (cr, 0, 0, w, h);
     cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, 0.4);
     cairo_fill (cr);
+    cairo_translate (cr, (w - len) / 2.0, (h - len) / 2.0);
+    cairo_scale (cr, len / 32.0, len / 32.0);
+    swfdec_gtk_player_draw_pause (cr);
   }
   cairo_destroy (cr);
   gdk_window_end_paint (player->target);
 }
 
 gboolean
-swfmoz_player_mouse_changed (SwfmozPlayer *player, int button, int x, int y, gboolean down)
+swfmoz_player_mouse_press (SwfmozPlayer *player, int x, int y, guint button)
 {
+  gboolean ret = FALSE;
+
   g_return_val_if_fail (SWFMOZ_IS_PLAYER (player), FALSE);
 
-  switch (button) {
-    case 1:
-      if (!swfdec_gtk_player_get_playing (SWFDEC_GTK_PLAYER (player->player))) {
-	if (!down)
-	  return FALSE;
-	swfdec_gtk_player_set_playing (SWFDEC_GTK_PLAYER (player->player), TRUE);
-      } else {
-	player->mouse_down = down;
-	plugin_push_allow_popups (player->instance, TRUE);
-	swfdec_player_handle_mouse (player->player, x, y, down ? 1 : 0);
-	plugin_pop_allow_popups (player->instance);
-      }
-      return TRUE;
-    case 3:
-      if (!down) {
-	swfmoz_player_popup_menu (player);
-	return TRUE;
-      }
-    default:
-      break;
+  if (button > 32)
+    return FALSE;
+
+  if (player->menu && GTK_WIDGET_VISIBLE (player->menu))
+    gtk_menu_popdown (GTK_MENU (player->menu));
+
+  if (swfdec_gtk_player_get_playing (SWFDEC_GTK_PLAYER (player->player))) {
+    plugin_push_allow_popups (player->instance, TRUE);
+    ret = swfdec_player_mouse_press (player->player, x, y, button);
+    plugin_pop_allow_popups (player->instance);
   }
-  return FALSE;
+  return ret;
 }
 
 gboolean
-swfmoz_player_mouse_moved (SwfmozPlayer *player, int x, int y)
+swfmoz_player_mouse_release (SwfmozPlayer *player, int x, int y, guint button)
+{
+  gboolean ret;
+
+  g_return_val_if_fail (SWFMOZ_IS_PLAYER (player), FALSE);
+
+  if (button > 32)
+    return FALSE;
+
+  if (swfdec_gtk_player_get_playing (SWFDEC_GTK_PLAYER (player->player))) {
+    plugin_push_allow_popups (player->instance, TRUE);
+    ret = swfdec_player_mouse_release (player->player, x, y, button);
+    plugin_pop_allow_popups (player->instance);
+  } else {
+    if (button == 1) {
+      swfdec_gtk_player_set_playing (SWFDEC_GTK_PLAYER (player->player), TRUE);
+      ret = TRUE;
+    } else {
+      ret = FALSE;
+    }
+  }
+
+  if (button == 3) {
+    swfmoz_player_popup_menu (player);
+    ret = TRUE;
+  }
+  return ret;
+}
+
+gboolean
+swfmoz_player_mouse_move (SwfmozPlayer *player, int x, int y)
 {
   g_return_val_if_fail (SWFMOZ_IS_PLAYER (player), FALSE);
 
   if (swfdec_gtk_player_get_playing (SWFDEC_GTK_PLAYER (player->player))) {
-    swfdec_player_handle_mouse (player->player, x, y, player->mouse_down ? 1 : 0);
+    swfdec_player_mouse_move (player->player, x, y);
   }
   return FALSE;
 }
